/*
 * Copyright (c) 2004 Stefan Ulrich
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * PAUL VOJTA OR ANY OTHER AUTHOR OF THIS SOFTWARE BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 */

/*
 * Panel 3 (Helper Applications) for xdvik preferences dialog.
 */

#include "xdvi-config.h"
#include "xdvi.h"

#include "x_util.h"
#include "xm_colorsel.h"
#include "topic-window.h"
#include "util.h"
#include "string-utils.h"
#include "message-window.h"
#include "events.h"
#include "xm_menu.h"

#include "xm_prefsP.h"
#include "xm_prefs.h"
#include "xm_prefs3.h"

#ifdef MOTIF /* entire file */

#include <X11/Xatom.h>

#include <Xm/Xm.h>
#include <Xm/Protocols.h>
#include <Xm/DialogS.h>
#include <Xm/LabelG.h>
#include <Xm/PushB.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/PanedW.h>
#include <Xm/DrawnB.h>
#include <Xm/RowColumn.h>
#include <Xm/Scale.h>
#include <Xm/ToggleBG.h>
#include <Xm/PushBG.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/FileSB.h>
#include <Xm/SelectioB.h>

struct choice_dialog_info {
    struct prefs_choice *prefs;
    Widget combo_box;
    Widget message_popup;
};

/* combo box only available with Motif 2.0;
 * the lesstif version is incomplete
 */
#if XmVERSION >= 2 && !defined(LESSTIF_VERSION)
#include <Xm/ComboBox.h>
#include <Xm/List.h>
#define USE_COMBOBOX 1
#else
#include <Xm/CascadeBG.h>
#define USE_COMBOBOX 0
#endif


/*
 * User clicked on `Help' in browser text input prompt
 */
static void
help_browser_dialog_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
    Widget popup = popup_message(XtParent(w),
				 MSG_HELP,
				 NULL,
				 "The browser is used by xdvi to browse remote documents. "
				 "Please enter the name of the browser (i.e. the executable program) "
				 "you want to use. The browser command may optionally contain a string "
				 "`%%s' which is replaced by the current URL. If no `%%s' is present, "
				 "the URL argument is simply appended to the command.");
    UNUSED(client_data);
    UNUSED(call_data);

    TRACE_GUI((stderr, "setting user data to %p", popup));
    XtVaSetValues(w, XmNuserData, popup, NULL);
}

/*
 * User clicked on `Help' in editor text input prompt
 */
static void
help_editor_dialog_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
    Widget popup = popup_message(XtParent(w),
				 MSG_HELP,
				 NULL,
				 "The editor is used by reverse search to open the TeX source for a DVI file "
				 "(see `Help' -> `Source Specials' for details)."
				 "Please enter the name of the editor executable you want to use. "
				 "The editor command may optionally contain two format strings: "
				 "`%%l' is replaced by the line number in the TeX file, and `%%f' by the file name. "
				 "(If the format strings are not present, they are appended implicitly.)");
    UNUSED(client_data);
    UNUSED(call_data);

    TRACE_GUI((stderr, "setting user data to %p", popup));
    XtVaSetValues(w, XmNuserData, popup, NULL);
}

/*
 * User cancelled browser or editor text input prompt
 */
static void
destroy_dialog_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
/*     Widget browser_combo = (Widget)client_data; */
    struct choice_dialog_info *info = (struct choice_dialog_info *)client_data;
    Widget browser_combo = NULL;
    Widget child;
    Widget browser_list;
    UNUSED(call_data);

    XtVaGetValues(w, XmNuserData, &child, NULL);
    TRACE_GUI((stderr, "GOT child: %p", child));

    if (info != NULL)
	browser_combo = info->combo_box;
    
    if (browser_combo != NULL) {
#if USE_COMBOBOX
	/* Need to unselect the `Other...' entry from the list; to do this,
	   we simply select the first item. */
	XtVaGetValues(browser_combo, XmNlist, &browser_list, NULL);
	if (browser_list == 0)
	    XDVI_ERROR((stderr, "couldn't get list component of combo box!\n"));
	else 
	    XmListSelectPos(browser_list, 1, True);
#else
	/*XXX: TODO */
#endif
    }

    /*
     * Also popdown dependent help window if it exists before destroying the widget;
     * otherwise, Motif may even crash in _XmIsFastSubclass() when the help window is moved!
     * 
     * Since `child' is the top-level xmDialogShellWidget, we need to get its
     * xmMessageBoxWidget child (by name ...)
    */
    if (child != NULL) {
	Widget message;
	if (get_widget_by_name(&message, child, Xdvi_MESSAGE_DIALOG_NAME, True))
	    XtCallCallbacks(message, XmNokCallback, NULL);
    }

    if (info != NULL) {
	remove_from_deplist(info->prefs, w);
	free(info);
    }
    
    /* destroy the parent of this dialog (the shell) */
    XtDestroyWidget(XtParent(w));
}

static void
h_get_input_wrapper(const char *listbox_name,
		    Widget w, XtPointer client_data, XtPointer call_data)
{
/*     Widget combobox = (Widget)client_data; */
    struct choice_dialog_info *info = (struct choice_dialog_info *)client_data;
    Widget combobox = info->combo_box;
    Widget child = info->message_popup;
    XmSelectionBoxCallbackStruct *cbs = (XmSelectionBoxCallbackStruct *)call_data;
    char *ptr;

    ASSERT(info != NULL, "client_data in h_get_input_wrapper mustn't be NULL!");
    XtVaGetValues(w, XmNuserData, &child, NULL);
    TRACE_GUI((stderr, "GOT child: %p", child));

    XmStringGetLtoR(cbs->value, G_charset, &ptr);
    if (strlen(ptr) == 0 || is_spaces_only(ptr)) {
	popup_message(XtParent(w),
		      MSG_ERR,
		      NULL,
		      "Empty input string");
	return;
    }
    
#if USE_COMBOBOX
    UNUSED(listbox_name);
    /* add user input to the combo box list, and make it current */
    XmComboBoxAddItem(combobox, cbs->value, 1, True);
    XmComboBoxSelectItem(combobox, cbs->value);
#else
    {
	Widget new_menu, parent, grandparent, rowcol;
	if ((parent = XtParent(w)) != NULL && (grandparent = XtParent(parent)) != NULL) {
	    /* add new item to front of list ... */
	    new_menu = XtVaCreateManagedWidget(ptr, xmPushButtonGadgetClass, grandparent,
					       XmNpositionIndex, 0,
					       NULL);
	    /* ... and make it current! */
	    if (get_widget_by_name(&rowcol, top_level, listbox_name, True))
		XtVaSetValues(rowcol, XmNmenuHistory, new_menu, NULL);
	}
    }
#endif

    /* as above: popdown help window */
    if (child != NULL) {
	Widget message;
	if (get_widget_by_name(&message, child, Xdvi_MESSAGE_DIALOG_NAME, True))
	    XtCallCallbacks(message, XmNokCallback, NULL);
    }

    remove_from_deplist(info->prefs, w);
    free(info);
    
    /* destroy the parent of this dialog (the shell) */
    XtDestroyWidget(XtParent(w));
}

static void
h_add_item_wrapper(const char *prompt_name,
		   const char *title_str, const char *label_str,
		   XtCallbackProc ok_cb, 
		   XtCallbackProc destroy_cb, 
		   XtCallbackProc help_cb, 
		   Widget w, XtPointer client_data, XtPointer call_data)
{
    struct prefs_choice *prefs = (struct prefs_choice *)client_data;
    struct choice_dialog_info *info = xmalloc(sizeof *info);
#if USE_COMBOBOX
    XmComboBoxCallbackStruct *cb = (XmComboBoxCallbackStruct *)call_data;
    char *choice = (char *)XmStringUnparse(cb->item_or_text, XmFONTLIST_DEFAULT_TAG,
					   XmCHARSET_TEXT, XmCHARSET_TEXT,
					   NULL, 0,
					   XmOUTPUT_ALL);
    if (cb->event == NULL) {
	fprintf(stderr, "Browsing: %s\n", choice);
    }
    else {
#else
	char *choice;
	XmString str;
	XtVaGetValues(w, XmNlabelString, &str, NULL);
	XmStringGetLtoR(str, G_charset, &choice);
#endif
	fprintf(stderr, "Widget: %p, Choice %s\n", w, choice);
	if (strcmp(choice, Xdvi_ADD_COMMAND_STR) == 0) {
	    Widget prompt_widget;
	    Arg args[8];
	    int n = 0;
	    XmString title = XmStringCreateLocalized((char *)title_str);
	    XmString label = XmStringCreateLocalized((char *)label_str);
	    
	    XtSetArg(args[n], XmNselectionLabelString, label); n++;
	    XtSetArg(args[n], XmNautoUnmanage, False); n++;
	    XtSetArg(args[n], XmNdialogTitle, title); n++;
	    XtSetArg(args[n], XmNuserData, NULL); n++;
	    prompt_widget = XmCreatePromptDialog(
#if USE_COMBOBOX
						 w,
#else
						 XtParent(w),
#endif
						 (char *)prompt_name, args, n);
	    XmStringFree(label);

	    add_to_deplist(prefs, prompt_widget);

	    info->prefs = prefs;
	    info->combo_box = w;
	    info->message_popup = NULL;
	    
	    /* Note: w is the browser_combo widget */
	    XtAddCallback(prompt_widget, XmNokCallback, ok_cb, (XtPointer)info);
	    XtAddCallback(prompt_widget, XmNcancelCallback, destroy_cb, (XtPointer)info);
	    XtAddCallback(prompt_widget, XmNhelpCallback, help_cb, (XtPointer)info);
	    
	    XtManageChild(prompt_widget);
	}
#if USE_COMBOBOX
    }
#endif
    XtFree (choice);
}

/*
 * User OK'ed browser text input prompt
 */
static void
get_browser_text_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
    h_get_input_wrapper(Xdvi_BROWSER_COMBO_NAME, w, client_data, call_data);
}


/*
 * User OK'ed editor text input prompt
 */
static void
get_editor_text_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
    h_get_input_wrapper(Xdvi_EDITOR_COMBO_NAME, w, client_data, call_data);
}


/*
 * User selected an item from browser combo box pulldown list
 */
static void
select_browser_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
    h_add_item_wrapper("browser_prompt",
		       "Xdvi: Add Browser Command",
		       "Browser Command (optional `%s' is replaced by URL): ",
		       get_browser_text_cb, destroy_dialog_cb, help_browser_dialog_cb,
		       w, client_data, call_data);
}

/*
 * User selected an item from editor combo box pulldown list
 */
static void
select_editor_cb(Widget w, XtPointer client_data, XtPointer call_data)
{
    UNUSED(client_data);
    h_add_item_wrapper("editor_prompt",
		       "Xdvi: Add Editor Command",
		       "Editor Command (optional: `%l' = line number, `%f' = file name): ",
		       get_editor_text_cb, destroy_dialog_cb, help_editor_dialog_cb,
		       w, client_data, call_data);
}

static Widget
h_create_command(const char *name, const char *label,
		 Widget parent, Widget top,
		 const char **command_list,
		 XtCallbackProc select_cb, struct prefs_choice *prefs)
{
#if USE_COMBOBOX
    XmStringTable str_list;
    Widget text_label, combo_box;
    size_t i, num;

    text_label = XmCreateLabelGadget(parent, (char *)label, NULL, 0);
    XtVaSetValues(text_label,
		  XmNtopAttachment, XmATTACH_WIDGET,
		  XmNtopWidget, top,
		  XmNleftAttachment, XmATTACH_FORM,
/* 		  XmNrightAttachment, XmATTACH_FORM, */
		  NULL);

    /* count elements in command_list */
    for (num = 0; command_list[num] != NULL; num++) { ; }
    
    str_list = (XmStringTable)XtMalloc(num * sizeof(XmString *));

    for (i = 0; i < num; i++) {
	if (strcmp(command_list[i], Xdvi_ADD_COMMAND_STR) == 0)
	    str_list[i] = XmStringCreateLtoR((char *)command_list[i], "MARKED");
	else
	    str_list[i] = XmStringCreateLtoR((char *)command_list[i], "UNMARKED");
    }
    combo_box = XtVaCreateWidget(name, xmComboBoxWidgetClass,
				 parent,
				 XmNtopAttachment, XmATTACH_WIDGET,
				 XmNtopWidget, top, /* if top == NULL, this used XmATTACH_FORM */
				 XmNleftAttachment, XmATTACH_WIDGET,
				 XmNleftWidget, text_label,
				 XmNrightAttachment, XmATTACH_FORM,
				 XmNcomboBoxType, XmDROP_DOWN_LIST,
				 XmNitems, str_list,
				 XmNitemCount, num,
				 NULL);
    for (i = 0; i < num; i++)
	XmStringFree(str_list[i]);
    XtFree((XtPointer)str_list);
    /*     XmComboBoxSelectItem(combo_box, str_list[0]); */

    fprintf(stderr, "combo box: %p\n", combo_box);
    XtAddCallback(combo_box, XmNselectionCallback, select_cb, (XtPointer)prefs); 
    
    /*
     * workaround for pointer grabbing bug (see xm_menu.c). We need to use
     * the popdownCallback of the internal `GrabShell' child of the combo box.
     */
    {
	Widget grab_shell;
	if (get_widget_by_name(&grab_shell, combo_box, "GrabShell", True))
	    XtAddCallback(grab_shell, XtNpopdownCallback, popdown_callback, NULL);
    }
    adjust_heights(text_label, combo_box, NULL);

    XtManageChild(text_label);
    XtManageChild(combo_box);

    return combo_box;
#else
    Widget menu, cascade;
    XmString str;
    Arg args[8];
    int n;
    size_t i;

    menu = XmCreatePulldownMenu(parent, "_pulldown", NULL, 0);
/*     XtVaSetValues(menu, */
/* 		  XmNtopAttachment, XmATTACH_WIDGET, */
/* 		  XmNtopWidget, top, */
/* 		  XmNleftAttachment, XmATTACH_FORM, */
/* 		  XmNrightAttachment, XmATTACH_FORM, */
/* 		  NULL); */

    str = XmStringCreateLocalized((char *)label);
    n = 0;
    XtSetArg(args[n], XmNsubMenuId, menu); n++;
    XtSetArg(args[n], XmNlabelString, str); n++;
    cascade = XmCreateOptionMenu(parent, (char *)name, args, n);
    fprintf(stderr, "cascade: %p, |%s|\n", cascade, name);
    if (top)
	XtVaSetValues(cascade,
		      XmNtopAttachment, XmATTACH_WIDGET,
		      XmNtopWidget, top,
		      XmNleftAttachment, XmATTACH_FORM,
		      NULL);
    else
	XtVaSetValues(cascade,
		      XmNtopAttachment, XmATTACH_FORM,
		      XmNleftAttachment, XmATTACH_FORM,
		      NULL);
    XmStringFree(str);

    for (i = 0; command_list[i] != NULL; i++) {
	Widget w;
	fprintf(stderr, "menu: %p\n", menu);
	w = XtVaCreateManagedWidget(command_list[i], xmPushButtonGadgetClass, menu, NULL);
	XtAddCallback(w, XmNactivateCallback, select_cb, menu); 
    }
    XtManageChild(cascade);
    return cascade;
#endif
}

Widget
prefs_helpers(struct topic_info *info)
{
    struct prefs_choice *prefs = (struct prefs_choice *)info->data;
    Widget form;
    Widget browser_command, editor_command;

    const char *browser_list[] = {
	"mozilla -remote 'openURL(%s,new-window)'",
	"netscape -raise -remote 'openURL(%s,new-window)'",
	"xterm -e lynx %s",
	Xdvi_ADD_COMMAND_STR,
	NULL /* terminate */
    };
    const char *editor_list[] = {
	"gnuclient -q +%l %f",
	"emacsclient --no-wait +%l %f",
	"gvim --servername xdvi --remote +%l %f",
	"nc -noask +%l %f",
	"xterm -e vi +%l %f",
	Xdvi_ADD_COMMAND_STR,
	NULL /* terminate */
    };
    
    form = XtVaCreateWidget("helpers_form", xmFormWidgetClass,
			    info->right_form,
			    XmNtopAttachment, XmATTACH_FORM,
			    XmNleftAttachment, XmATTACH_FORM,
			    XmNrightAttachment, XmATTACH_FORM,
			    XmNbottomAttachment, XmATTACH_FORM,
			    XmNverticalSpacing, 10,
			    XmNhorizontalSpacing, 10,
			    NULL);
    
    browser_command = h_create_command(Xdvi_BROWSER_COMBO_NAME, "Web Browser: ",
				       form, NULL,
				       browser_list,
				       select_browser_cb, prefs);
    
    editor_command = h_create_command(Xdvi_EDITOR_COMBO_NAME, "Editor for Source Specials: ",
				      form, browser_command,
				      editor_list,
				      select_editor_cb, prefs);
    
/* #if PS_GS */
/*     Widget b2 = XmCreateLabelGadget(rowcol, "[x] Use Ghostscript to interpret PS specials", NULL, 0); */

    /* TODO: don't need this??? Similar: ps2pdf, dvips conversion?? */
    /*     Widget b3 = XmCreateLabelGadget(rowcol, "Path to Ghostscript: ____________", NULL, 0); */
    
/* #endif */
/*     Widget b4 = XmCreateLabelGadget(rowcol, "Editor for source specials: <pulldown> [Other ...]", NULL, 0); */
/*     XtManageChild(b1); */
/*     XtManageChild(b2); */
/*     XtManageChild(b3); */
/*     XtManageChild(b4); */
/*     XtManageChild(rowcol); */

    
    return form;
}

#endif /* MOTIF */

