#include "xdvi-config.h"

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>


#if !MOTIF
# include <X11/Xaw/Form.h>
#endif

#include "x_util.h"
#include "util.h"
#include "statusline.h"
#include "message-window.h"
#include "events.h" /* for set_bar_value() */

#define DEBUG_SCROLL_IF_NEEDED 0

Boolean do_autoscroll = False;


/* Center window wa over window wb */
void
center_window(Widget wa, Widget wb)
{
    Position x, y;
    Dimension w1, h1, w2, h2;

    if (!XtIsRealized(wa) || !XtIsRealized(wb))
	return;
    
    XtVaGetValues(wa,
		  XtNwidth, &w1,
		  XtNheight, &h1,
		  NULL);
    XtVaGetValues(wb,
		  XtNwidth, &w2,
		  XtNheight, &h2,
		  XtNx, &x,
		  XtNy, &y,
		  NULL);
    XtVaSetValues(wa, XtNx, x + (w2 - w1) / 2,
		  XtNy, y + (h2 - h1) / 2,
		  NULL);
}

/* Position window w at coordinates x, y */
void
position_window(Widget w, Position x, Position y)
{
    if (!XtIsRealized(w))
	return;

    TRACE_GUI((stderr, "positioning %ld at %d, %d", (unsigned long)w, x, y));
    XtVaSetValues(w, XtNx, x, XtNy, y, NULL);
}

/*
  Used for the hyperref and forward search markers: scroll the page
  to make the marker visible.
*/
void
scroll_page_if_needed(int x_min, int x_max, int y_min, int y_max)
{
    Position drawing_x, drawing_y, drawing_h, clip_x, clip_y, clip_h, clip_w;
    int test_scroll, need_v_scroll = 0, need_h_scroll = 0;

    if (!do_autoscroll)
	return;
    
    XtVaGetValues(clip_widget,
		  XtNx, &clip_x, XtNy, &clip_y,
		  XtNheight, &clip_h, XtNwidth, &clip_w,
		  NULL);
    XtVaGetValues(draw_widget,
		  XtNx, &drawing_x, XtNy, &drawing_y,
		  XtNheight, &drawing_h,
		  NULL);

#if DEBUG_SCROLL_IF_NEEDED
    fprintf(stderr, "y: %d, drawing_y: %d, clip_h: %d\n", y, drawing_y, clip_h);
#endif
    /* check if we need to scroll vertically; first, down for y_min */
    test_scroll = y_min + drawing_y - clip_h;
    if ((expert_flag & XPRT_SHOW_STATUSLINE) != 0)
	test_scroll += global_statusline_h;

    TRACE_SRC((stderr, "test_scroll vertically: %d", test_scroll));
#if DEBUG_SCROLL_IF_NEEDED
    fprintf(stderr, "%d + %d > %d?\n", drawing_y, y_min, clip_h);
#endif
    if (test_scroll > 0) { /* need to scroll down? */
	need_v_scroll = test_scroll;
	TRACE_SRC((stderr, "need_v_scroll down: %d", need_v_scroll));
    }
    else if (abs(drawing_y) + 2 > y_max) { /* need to scroll up? */
	need_v_scroll = -((abs(drawing_y) - y_max) + 1);
	TRACE_SRC((stderr, "need_v_scroll up: %d (%d > %d; %d)", need_v_scroll, abs(drawing_y), y_max, clip_y));
    }

    /* check if we need to scroll horizontally; x_min < 0 blocks this
       (e.g. for hyperref, where we don't want it) */
    if (x_min >= 0) {
	test_scroll = x_min + drawing_x - clip_w + 1;
	TRACE_SRC((stderr, "test_scroll horizontally: %d", test_scroll));
	
	if (test_scroll > 0) {
	    /* need to scroll to right (i.e. make stuff on right-hand side visible)? */
	    need_h_scroll = test_scroll;
	    TRACE_SRC((stderr, "need_h_scroll right: %d", need_h_scroll));
	}
	else if (abs(drawing_x) > x_max) {
	    /* need to scroll to left (i.e. make stuff on left-hand side visible)? */
	    need_h_scroll =  -(abs(drawing_x) - x_max);
	    TRACE_SRC((stderr, "need_h_scroll left: %d", need_h_scroll));
	}
    }

    /* FIXME: should we not scroll if keep_flag is active? */
    if (need_v_scroll != 0 && y_bar != NULL) {
#ifdef MOTIF
	/* need to add new value to current one */
	XtVaGetValues(y_bar, XmNvalue, &test_scroll, NULL);
	(void)set_bar_value(y_bar, test_scroll + need_v_scroll, (int)(page_h - mane.height));
#else
	XtCallCallbacks(y_bar, XtNscrollProc, (XtPointer)need_v_scroll);
#endif
    }

    if (need_h_scroll != 0 && x_bar != NULL) {
#ifdef MOTIF
	/* need to add new value to current one */
	XtVaGetValues(x_bar, XmNvalue, &test_scroll, NULL);
	(void)set_bar_value(x_bar, test_scroll + need_h_scroll, (int)(page_w - mane.width));
#else
	XtCallCallbacks(x_bar, XtNscrollProc, (XtPointer)need_h_scroll);
#endif
    }

    do_autoscroll = False;
}

void adjust_width(Widget a, Widget b)
{
    Dimension w1, w2;
    
    XtVaGetValues(a, XtNwidth, &w1, NULL);
    XtVaGetValues(b, XtNwidth, &w2, NULL);
    
    if (w1 < w2)
	XtVaSetValues(a, XtNwidth, w2, NULL);
    else if (w2 > w1)
	XtVaSetValues(b, XtNwidth, w1, NULL);
}

/* change a GC and return it, or create one if it doesn't exist yet */
GC
set_or_make_gc(GC gc, int function, Pixel fg, Pixel bg)
{
    XGCValues values;
    values.function = function;
    values.foreground = fg;
    values.background = bg;

    /* Since print is in round dots we make drawings as "smooth" as possible. */
    values.cap_style = CapRound;
    values.join_style = JoinRound;

    if (gc != NULL)
	XChangeGC(DISP, gc, GCFunction | GCForeground | GCBackground
		  | GCCapStyle | GCJoinStyle, &values);
    else
	gc = XCreateGC(DISP, XtWindow(top_level), GCFunction | GCForeground | GCBackground
		       | GCCapStyle | GCJoinStyle, &values);

    return gc;
}

/*
 *	Atom handling.
 */

static char *atom_names[] = {
    "XDVI_WINDOWS",
    "DVI_NAME",
    "SRC_GOTO",
    "RELOAD",
    "NEWDOC",
    "NEWPAGE",
    "RAISE",
    "FIND_STRING"
};
static Atom atoms[XtNumber(atom_names)];

Window
get_window_id(char *window_p)
{
    Window w;
    
#ifndef WORD64
    w = (*((xuint32 *) window_p));
#else
# if WORDS_BIGENDIAN
    w = ((unsigned long)wp[0] << 24) | ((unsigned long)wp[1] << 16)
	| ((unsigned long)wp[2] << 8) | (unsigned long)wp[3];
# else
    w = ((unsigned long)wp[3] << 24) | ((unsigned long)wp[2] << 16)
	| ((unsigned long)wp[1] << 8) | (unsigned long)wp[0];
# endif
#endif
    return w;
}

size_t
property_get_data(Window w, Atom a, char **ret_buf,
		  int (*x_get_property)(Display *, Window, Atom, long,
					long, Bool, Atom, Atom *, int *, unsigned long *,
					unsigned long *, unsigned char **))
{
    /* all of these are in 8-bit units */
    unsigned long byte_offset = 0;
    Atom type_ret;
    int format_ret;
    unsigned long nitems_ret;
    unsigned long bytes_after_ret = 0;
    unsigned char *prop_ret = NULL;
    
    /*
     * buffer for collecting returned data; this is static to
     * avoid expensive malloc()s at every call (which is often!)
     */
    static unsigned char *buffer = NULL;
    static size_t buffer_len = 0;

    while (x_get_property(DISP, w,
			  a, byte_offset / 4, (bytes_after_ret + 3) / 4, False,
			  a, &type_ret, &format_ret, &nitems_ret,
			  &bytes_after_ret, &prop_ret)
	   == Success) {

	if (type_ret != a || format_ret == 0)
	    break;

	nitems_ret *= (format_ret / 8);	/* convert to bytes */

	while ((byte_offset + nitems_ret) >= buffer_len) {
	    buffer_len += 256;
	    buffer = xrealloc(buffer, buffer_len);
	}

	/* the +1 captures the extra '\0' that Xlib puts after the end.  */
	memcpy(buffer + byte_offset, prop_ret, nitems_ret + 1);
	byte_offset += nitems_ret;

	XFree(prop_ret);
	prop_ret = NULL;

	if (bytes_after_ret == 0)	/* got all data */
	    break;
    }

    if (prop_ret != NULL)
	XFree(prop_ret);

    *ret_buf = (char *)buffer;
    return byte_offset;
}

size_t
property_get_window_list(char **window_list)
{
    size_t len = property_get_data(DefaultRootWindow(DISP),
				   atom_xdvi_windows(), window_list,
				   XGetWindowProperty);
    if (len == 0) {
	TRACE_CLIENT((stderr, "No \"xdvi windows\" property found"));
	return 0;
    }
    
    if (len % 4 != 0) {
	TRACE_CLIENT((stderr, "\"XDVI_WINDOWS\" property had incorrect size; deleting it."));
	XDeleteProperty(DISP, DefaultRootWindow(DISP), atom_xdvi_windows());
	return 0;
    }
    return len;
}

void
set_window_id(Window w, unsigned char *data)
{
#if WORDS_BIGENDIAN
    data[0] = (unsigned int)w >> 24;
    data[1] = (unsigned int)w >> 16;
    data[2] = (unsigned int)w >> 8;
    data[3] = (unsigned int)w;
#else
    data[0] = (unsigned int)w;
    data[1] = (unsigned int)w >> 8;
    data[2] = (unsigned int)w >> 16;
    data[3] = (unsigned int)w >> 24;
#endif
}

/**
 **	set_dvi_property sets the appropriate property for the main
 **	window (used in source special handoff).
 **/

void
set_dvi_property(void)
{
    XChangeProperty(DISP, XtWindow(top_level), atom_dvi_file(), atom_dvi_file(),
		    8, PropModeReplace, (unsigned char *)dvi_property, dvi_property_length);
}


/*
 * Delete all occurences of window w from the window list property. Then,
 * if `prepend' is true, prepend the window ID to the existing list.
 */
void
update_window_property(Window w, Boolean prepend)
{
    char *window_list;
    size_t window_list_len;
    char *window_list_end;
    char *wp;


    /* this allocates window_list */
    if ((window_list_len = property_get_window_list(&window_list)) == 0)
	return;

    /* Loop over list of windows.  */
    window_list_end = window_list + window_list_len;

#if 0
    if (prepend && window_list_len == 4) /* if ours is the only window */
	return;
#endif /* 0 */
    
    for (wp = window_list; wp < window_list_end; wp += 4) {
	if (get_window_id(wp) == w) { /* match, remove our ID */
	    window_list_len -= 4;
	    window_list_end -= 4;
	    memmove(wp, wp + 4, window_list_end - wp);

	    if (prepend) { /* add our ID again to front */
#ifdef WORD64
		unsigned char data[4];
		set_window_id(w, data);
#else
		xuint32 data = w;
#endif
/* 		if (debug & DBG_CLIENT) { */
/* 		    unsigned char *wp; */
/* 		    int i; */
/* 		    fprintf(stderr, "======== BEFORE replacing:\n"); */
/* 		    for (i = 0, wp = window_list; wp < window_list_end; wp += 4, i++) { */
/* 			Window w = get_window_id(wp); */
/* 			fprintf(stderr, "Window %d: %08lx\n", i, w); */
/* 		    } */
/* 		} */
		/* Note: no need to realloc window_list, since the original length
		   was sufficient for all elements.
		*/
		window_list_len += 4;
		memmove(window_list + 4, window_list, window_list_len - 4);
		memcpy(window_list, &data, 4);
/* 		if (debug & DBG_CLIENT) { */
/* 		    unsigned char *wp; */
/* 		    int i; */
/* 		    fprintf(stderr, "======== after replacing:\n"); */
/* 		    for (i = 0, wp = window_list; wp < window_list_end; wp += 4, i++) { */
/* 			Window w = get_window_id(wp); */
/* 			fprintf(stderr, "Window %d: %08lx\n", i, w); */
/* 		    } */
/* 		} */
	    }
	    
	    if (window_list_len == 0)
		XDeleteProperty(DISP, DefaultRootWindow(DISP),
				atom_xdvi_windows());
	    else
		XChangeProperty(DISP, DefaultRootWindow(DISP),
				atom_xdvi_windows(), atom_xdvi_windows(), 32,
				PropModeReplace, (unsigned char *)window_list,
				window_list_len / 4);

	    XFlush(DISP);
	    return;
	}
    }
    TRACE_CLIENT((stderr, "Couldn't find window %08lx in list when exiting.", w));

    return;
}

void
property_initialize(void)
{
    if (
#if XlibSpecificationRelease >= 6
	!XInternAtoms(DISP, atom_names, XtNumber(atom_names), False, atoms)
#else
	(atoms[0] = XInternAtom(DISP, atom_names[0], False)) == None
	|| (atoms[1] = XInternAtom(DISP, atom_names[1], False)) == None
	|| (atoms[2] = XInternAtom(DISP, atom_names[2], False)) == None
	|| (atoms[3] = XInternAtom(DISP, atom_names[3], False)) == None
	|| (atoms[4] = XInternAtom(DISP, atom_names[4], False)) == None
	|| (atoms[5] = XInternAtom(DISP, atom_names[5], False)) == None
	|| (atoms[6] = XInternAtom(DISP, atom_names[6], False)) == None
	|| (atoms[6] = XInternAtom(DISP, atom_names[7], False)) == None
#endif
	) {
	XDVI_FATAL((stderr, "XtInternAtoms failed."));
    }

    if (debug & DBG_CLIENT) {
	Cardinal k;
	for (k = 0; k < XtNumber(atom_names); ++k)
	    TRACE_CLIENT((stderr, "Atom(%s) = %lu", atom_names[k], atoms[k]));
    }
}


Atom atom_xdvi_windows(void)
{
    return atoms[0];
}

Atom atom_dvi_file(void)
{
    return atoms[1];
}

Atom atom_src_goto(void)
{
    return atoms[2];
}

Atom atom_reload(void)
{
    return atoms[3];
}

Atom atom_newdoc(void)
{
    return atoms[4];
}

Atom atom_newpage(void)
{
    return atoms[5];
}

Atom atom_raise(void)
{
    return atoms[6];
}

Atom atom_find_string(void)
{
    return atoms[7];
}

/*
 * Syntetisize a button press object for the `button' passed as
 * the second argument.
 */
void
synthetisize_event(XEvent *ev, Widget button)
{
    memset(ev, 0, sizeof(XButtonPressedEvent));
    ev->type = ButtonPress;
    ev->xbutton.serial = 1;
    ev->xbutton.send_event = True;
    ev->xbutton.button = 1;
    ev->xbutton.display = XtDisplayOfObject(button);
    ev->xbutton.window = XtWindowOfObject(button);
}

#ifdef MOTIF
int xm_get_height(Widget w)
{
    int ret_h = 0;
    static Arg args = { XtNheight, (XtArgVal)0 };
    ASSERT(w != NULL, "widget in xm_get_width mustn't be NULL!");

    args.value = (XtArgVal)&ret_h;
    XtGetValues(w, &args, 1);
    TRACE_GUI((stderr, "xm_get_height: %d", ret_h));
    return ret_h;
}

int xm_get_width(Widget w)
{
    int ret_w = 0;
    static Arg args = { XtNwidth, (XtArgVal)0 };
    ASSERT(w != NULL, "widget in xm_get_width mustn't be NULL!");

    args.value = (XtArgVal)&ret_w;
    XtGetValues(w, &args, 1);
    TRACE_GUI((stderr, "xm_get_width: %d", ret_w));
    return ret_w;
}

/*
 * Get pixel from color `colorname'. We try to keep this as
 * high-level as possible, with simple black as fallback.
*/
void
str_to_pixel(Widget w, const char *colorname, Pixel *ret)
{
    XrmValue from, to;

    from.addr = (char *)colorname;
    from.size = strlen(from.addr) + 1;
    to.addr = (XtPointer)ret;
    to.size = sizeof(Pixel);
    if (!XtConvertAndStore(w, XtRString, &from, XtRPixel, &to)) {
	fprintf(stderr, "String to pixel conversion failed for %s!\n", colorname);
	from.addr = (char *)"black";
	from.size = strlen(from.addr) + 1;
	to.addr = (XtPointer)ret;
	to.size = sizeof(Pixel);
	XtConvertAndStore(w, XtRString, &from, XtRPixel, &to);
    }
}

/*
 * Convert pixel value `pix' to color name passed as str,
 * of length len, or `black' if conversion failed.
 * -- FIXME: This is broken!!!
*/
void
pixel_to_str(Widget w, Pixel pix, char *str, size_t len)
{
    XrmValue from, to;

    from.addr = (XtPointer)&pix;
    from.size = sizeof(Pixel);
    to.addr = str;
    to.size = len;

    if (!XtConvertAndStore(w, XtRString, &from, XtRPixel, &to)) {
	fprintf(stderr, "Pixel to String conversion failed for %ld!\n", pix);
	sprintf(str, "white");
    }
}

/* Free color, and initialize it anew with pixel value `pix' */
void
pixel_to_color(Pixel pix, XColor *color, Display *display, Colormap colormap)
{
    XColor test;
    test.pixel = pix;
    XQueryColor(display, colormap, &test);

    color->red = test.red;
    color->green = test.green;
    color->blue = test.blue;
    
    if (!XAllocColor(display, colormap, color)) {
	fprintf(stderr, "Fatal: Couldn't XAllocColor!");
	exit(1);
    }
}
#endif /* MOTIF */

/* helper routine for get_matching_parent() */
static Widget
matches_parent(Widget w, const char *name)
{
    for (; w != NULL; w = XtParent(w)) {
	char *ptr;
	ptr = XtName(w);
	TRACE_GUI((stderr, "parent: %s", ptr == NULL ? "<NULL>" : ptr));
	if (ptr != NULL && strcmp(ptr, name) == 0) { /* found */
	    TRACE_GUI((stderr, "match!"));
	    break;
	}
    }
    TRACE_GUI((stderr, "returning: %p (0x%lx)", w, w ? XtWindow(w) : 0));
    return w;
}

/* Traverse widget hieararchy upwards until a widget matches
   a name in the (NULL-terminated) list fmt, or Widget d (`default')
   if none is found.
*/
Widget
get_matching_parent(Widget w, Widget d, const char *fmt, ...)
{
    Widget parent = d;
    const char *str = fmt;

    va_list argp;
    va_start(argp, fmt);

    TRACE_GUI((stderr, "get_matching_parent of %p (0x%lx)", w, XtWindow(w)));
    while (str != NULL) {
	Widget p;
	if ((p = matches_parent(w, str)) != NULL) {
	    parent = p;
	    break;
	}
	str = va_arg(argp, char *);
    }
	
    va_end(argp);

    return parent;
}

/* Return True if p is a parent of widget w, stopping (and returning FALSE)
   if s is reached (which should be the toplevel window of the hierarchy).
*/
Boolean
widget_is_parent(Widget w, Widget p, Widget s)
{
    Widget curr = XtParent(w);
    while (curr != NULL && curr != s) {
	fprintf(stderr, "Comparing: %p - %p\n", curr, p);
	if (curr == p)
	    return True;
	curr = XtParent(curr);
    }
    return False;
}

/*
  Adjust height in a NULL-terminated list of widgets. For Motif,
  this works better than the following adjust_vertically().
*/
void
adjust_heights(Widget w, ...)
{
    va_list ap;
    Widget curr;
    Dimension h, max;

    ASSERT(w != NULL, "Must have at least one element in va_list for adjust_heights!");

#if MOTIF
#define HEIGHT XmNheight
#else
#define HEIGHT XtNheight
#endif
    
    /* initialize maximum */
    XtVaGetValues(w, HEIGHT, &max, NULL);

    /* get maximum height */
    va_start(ap, w);
    while ((curr = va_arg(ap, Widget)) != NULL) {
	XtVaGetValues(curr, HEIGHT, &h, NULL);
	if (h > max)
	    max = h;
    }
    va_end(ap);

    /* set maximum height for all widgets */
    XtVaSetValues(w, HEIGHT, max, NULL);
    
    va_start(ap, w);
    while ((curr = va_arg(ap, Widget)) != NULL)
	XtVaSetValues(curr, HEIGHT, max, NULL);
    va_end(ap);

#undef HEIGHT

}

/*
  Adjust height in a NULL-terminated list of widgets. For Motif,
  this works better than the following adjust_vertically().
*/
void
adjust_heights_min(Widget w, ...)
{
    va_list ap;
    Widget curr;
    Dimension h, min;

    ASSERT(w != NULL, "Must have at least one element in va_list for adjust_heights!");

#if MOTIF
#define HEIGHT XmNheight
#else
#define HEIGHT XtNheight
#endif
    
    /* initialize minimum */
    XtVaGetValues(w, HEIGHT, &min, NULL);

    /* get minimum height */
    va_start(ap, w);
    while ((curr = va_arg(ap, Widget)) != NULL) {
	XtVaGetValues(curr, HEIGHT, &h, NULL);
	if (h < min)
	    min = h;
    }
    va_end(ap);

    /* set maximum height for all widgets */
    XtVaSetValues(w, HEIGHT, min, NULL);
    
    va_start(ap, w);
    while ((curr = va_arg(ap, Widget)) != NULL)
	XtVaSetValues(curr, HEIGHT, min, NULL);
    va_end(ap);

#undef HEIGHT

}

/* adjust two widgets vertically */
void
adjust_vertically(Widget w1, Widget w2, int default_dist)
{
    Dimension h1, h2;
#if MOTIF
    XtVaGetValues(w1, XmNheight, &h1, NULL);
    XtVaGetValues(w2, XmNheight, &h2, NULL);
    XtVaSetValues(w1, XmNtopOffset, default_dist + (h2 - h1) / 2, NULL);
    /*     XtVaSetValues(w2, XmNtopOffset, default_dist + (h2 - h1) / 2, NULL); */
#else
    XtVaGetValues(w1, XtNheight, &h1, NULL);
    XtVaGetValues(w2, XtNheight, &h2, NULL);
    XtVaSetValues(w1, XtNvertDistance, default_dist + (h2 - h1) / 2, NULL);
    XtVaSetValues(w2, XtNvertDistance, default_dist + (h2 - h1) / 2, NULL);
#endif /* MOTIF */
}


/*
 * This is a hack to block further processing of some events on widgets:
 * Add as an event handler for all mouse/key events for a specific widget.
 */
void
block_event_callback(Widget w, XtPointer client_data,
		     XEvent *ev, Boolean *cont)
{
    UNUSED(w);
    UNUSED(client_data);
    UNUSED(ev);
    
    /* Don't propagate this event further down... */
    *cont = False;

    return;
}


/* This is incredibly silly ... programming using Emacs keyboard macros */
static void
copy_resource_pointers(const struct x_resources *source, struct x_resources *target)
{
    target->main_translations = source->main_translations ? xstrdup(source->main_translations) : NULL;
    target->wheel_translations = source->wheel_translations ? xstrdup(source->wheel_translations) : NULL;
    target->tick_units = source->tick_units ? xstrdup(source->tick_units) : NULL;
    target->sidemargin = source->sidemargin ? xstrdup(source->sidemargin) : NULL;
    target->topmargin = source->topmargin ? xstrdup(source->topmargin) : NULL;
    target->xoffset = source->xoffset ? xstrdup(source->xoffset) : NULL;
    target->yoffset = source->yoffset ? xstrdup(source->yoffset) : NULL;
    target->paper = source->paper ? xstrdup(source->paper) : NULL;
    target->alt_font = source->alt_font ? xstrdup(source->alt_font) : NULL;
    target->mfmode = source->mfmode ? xstrdup(source->mfmode) : NULL;
    target->editor = source->editor ? xstrdup(source->editor) : NULL;
    target->src_pos = source->src_pos ? xstrdup(source->src_pos) : NULL;
    target->find_string = source->find_string ? xstrdup(source->find_string) : NULL;
    target->text_encoding = source->text_encoding ? xstrdup(source->text_encoding) : NULL;
    target->sub_pixels = source->sub_pixels ? xstrdup(source->sub_pixels) : NULL;
#ifdef VMS
    target->fore_color = source->fore_color ? xstrdup(source->fore_color) : NULL;
    target->back_color = source->back_color ? xstrdup(source->back_color) : NULL;
#endif
    target->icon_geometry = source->icon_geometry ? xstrdup(source->icon_geometry) : NULL;
    target->pause_special = source->pause_special ? xstrdup(source->pause_special) : NULL;
    target->gs_path = source->gs_path ? xstrdup(source->gs_path) : NULL;
    target->gs_palette = source->gs_palette ? xstrdup(source->gs_palette) : NULL;
#ifdef MAGICK
    target->magick_cache = source->magick_cache ? xstrdup(source->magick_cache) : NULL;
#endif
    target->debug_arg = source->debug_arg ? xstrdup(source->debug_arg) : NULL;
    target->menu_translations = source->menu_translations ? xstrdup(source->menu_translations) : NULL;
#ifdef MOTIF
    target->toolbar_translations = source->toolbar_translations ? xstrdup(source->toolbar_translations) : NULL;
    target->toolbar_pixmap_file = source->toolbar_pixmap_file ? xstrdup(source->toolbar_pixmap_file) : NULL;
#endif
    target->mg_arg[0] = source->mg_arg[0] ? xstrdup(source->mg_arg[0]) : NULL;
    target->mg_arg[1] = source->mg_arg[1] ? xstrdup(source->mg_arg[1]) : NULL;
    target->mg_arg[2] = source->mg_arg[2] ? xstrdup(source->mg_arg[2]) : NULL;
    target->mg_arg[3] = source->mg_arg[3] ? xstrdup(source->mg_arg[3]) : NULL;
    target->mg_arg[4] = source->mg_arg[4] ? xstrdup(source->mg_arg[4]) : NULL;
    target->dvips_path = source->dvips_path ? xstrdup(source->dvips_path) : NULL;
    target->ps2pdf_path = source->ps2pdf_path ? xstrdup(source->ps2pdf_path) : NULL;
    target->dvips_printer_str = source->dvips_printer_str ? xstrdup(source->dvips_printer_str) : NULL;
    target->dvips_options_str = source->dvips_options_str ? xstrdup(source->dvips_options_str) : NULL;
    target->rule_color = source->rule_color ? xstrdup(source->rule_color) : NULL;
    target->link_color = source->link_color ? xstrdup(source->link_color) : NULL;
    target->visited_link_color = source->visited_link_color ? xstrdup(source->visited_link_color) : NULL;
    target->browser = source->browser ? xstrdup(source->browser) : NULL;
    target->CMD_WWWbrowser = source->CMD_WWWbrowser ? xstrdup(source->CMD_WWWbrowser) : NULL;
    target->unknown_mime_suffix = source->unknown_mime_suffix ? xstrdup(source->unknown_mime_suffix) : NULL;
    target->no_mime_suffix = source->no_mime_suffix ? xstrdup(source->no_mime_suffix) : NULL;
    target->anchor_pos = source->anchor_pos ? xstrdup(source->anchor_pos) : NULL;
    target->help_general = source->help_general ? xstrdup(source->help_general) : NULL;
    target->help_hypertex = source->help_hypertex ? xstrdup(source->help_hypertex) : NULL;
    target->help_othercommands = source->help_othercommands ? xstrdup(source->help_othercommands) : NULL;
    target->help_pagemotion = source->help_pagemotion ? xstrdup(source->help_pagemotion) : NULL;
    target->help_marking = source->help_marking ? xstrdup(source->help_marking) : NULL;
    target->help_rulermode = source->help_rulermode ? xstrdup(source->help_rulermode) : NULL;
    target->help_search = source->help_search ? xstrdup(source->help_search) : NULL;
    target->help_mousebuttons = source->help_mousebuttons ? xstrdup(source->help_mousebuttons) : NULL;
    target->help_sourcespecials = source->help_sourcespecials ? xstrdup(source->help_sourcespecials) : NULL;
}

void
free_resources(struct x_resources *res)
{
    free_resource_pointers(res);
    free(res);
}

/* as above ... */
void
free_resource_pointers(struct x_resources *res)
{
    free((char *)res->main_translations);
    free((char *)res->wheel_translations);
    free((char *)res->tick_units);
    free((char *)res->sidemargin);
    free((char *)res->topmargin);
    free((char *)res->xoffset);
    free((char *)res->yoffset);
    free((char *)res->paper);
    free((char *)res->alt_font);
    free((char *)res->mfmode);
    free((char *)res->editor);
    free((char *)res->src_pos);
    free((char *)res->find_string);
    free((char *)res->text_encoding);
    free((char *)res->sub_pixels);
#ifdef VMS
    free((char *)res->fore_color);
    free((char *)res->back_color);
#endif
    free((char *)res->icon_geometry);
    free((char *)res->pause_special);
    free((char *)res->gs_path);
    free((char *)res->gs_palette);
#ifdef MAGICK
    free((char *)res->magick_cache);
#endif
    free((char *)res->debug_arg);
    free((char *)res->menu_translations);
#ifdef MOTIF
    free((char *)res->toolbar_translations);
    free((char *)res->toolbar_pixmap_file);
#endif
    free((char *)res->mg_arg[0]);
    free((char *)res->mg_arg[1]);
    free((char *)res->mg_arg[2]);
    free((char *)res->mg_arg[3]);
    free((char *)res->mg_arg[4]);
    free((char *)res->dvips_path);
    free((char *)res->ps2pdf_path);
    free((char *)res->dvips_printer_str);
    free((char *)res->dvips_options_str);
    free((char *)res->rule_color);
    free((char *)res->link_color);
    free((char *)res->visited_link_color);
    free((char *)res->browser);
    free((char *)res->CMD_WWWbrowser);
    free((char *)res->unknown_mime_suffix);
    free((char *)res->no_mime_suffix);
    free((char *)res->anchor_pos);
    free((char *)res->help_general);
    free((char *)res->help_hypertex);
    free((char *)res->help_othercommands);
    free((char *)res->help_pagemotion);
    free((char *)res->help_marking);
    free((char *)res->help_rulermode);
    free((char *)res->help_search);
    free((char *)res->help_mousebuttons);
    free((char *)res->help_sourcespecials);
}


/* XXX: TODO
   Also create a copy the global resource AT STARTUP, which can then
   be used to change settings (unlike current global resource).
*/
void
copy_resources(const struct x_resources *source, struct x_resources *target)
{
    ASSERT(source != NULL, "source mustn't be NULL!");
    ASSERT(target != NULL, "target mustn't be NULL!");
    
    /* copy the non-pointer data */
    memcpy(target, source, sizeof *source);
    /* copy all pointer data, using the ugly hacks from above */
    copy_resource_pointers(source, target);
    fprintf(stderr, "Editor command is: |%s|\n", target->editor);
}

Boolean
get_widget_by_name(Widget *ret, Widget parent, const char *name, Boolean report_error)
{
    char buf[1024];
    Widget test;

    if (strlen(name) > 1023) {
	popup_message(top_level,
		      MSG_ERR,
		      REPORT_XDVI_BUG_TEMPLATE,
		      "Widget name `%s' too long, couldn't get parent", name);
	return False;
    }
    sprintf(buf, "*%s", name);
    if ((test = XtNameToWidget(parent, buf)) != NULL) {
	*ret = test;
	return True;
    }
    else {
	if (report_error)
	    popup_message(top_level,
			  MSG_ERR,
			  REPORT_XDVI_BUG_TEMPLATE,
			  "XtNameToWidget failed for `%s', parent `%s'", name, XtName(parent));
	return False;
    }
}

void
unexpected_widget_in_callback(Widget w, const char *callback)
{
    ASSERT(w != NULL, "Widget mustn't be NULL!");
    popup_message(top_level,
		  MSG_ERR,
		  REPORT_XDVI_BUG_TEMPLATE,
		  "Unexpected widget `%s' in callback `%s'",
		  XtName(w), callback);
}


